/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.DataTypes;

import java.util.Hashtable;

import uk.ac.bbsrc.babraham.ChipMonk.Preferences.ChipMonkPreferences;


public class DataSet implements DataStore, Comparable {

	private Hashtable data = new Hashtable((int)(ChipMonkPreferences.getInstance().getApproxProbeSetSize()*1.33f),0.75f);
	private String name;
	private DataCollection collection = null;
	
	public DataSet (String name) {
		this.name = new String(name);
	}
	
	public void setName (String name) {
		this.name = name;
		if (collection != null) {
			collection.dataSetRenamed(this);
		}
	}
	
	public void setCollection (DataCollection collection) {
		this.collection = collection;
	}
	
	public void addData (Probe p, double signal, double control) {
		
		// We store the data in an array to save overhead.  The values
		// stored are:
		// 0 - raw signal
		// 1 - raw control
		// 2 - normalised signal
		// 3 - normalised control
		
		// Because we're often working with log data it's a real problem
		// if we get a signal or double value which is exactly 0.  We therefore
		// set a minimum value which these values can have.  We also avoid
		// divide by 0 errors when calculating ratios.
		
		if (signal < DataStore.LOWEST_ALLOWED_VALUE)
			signal = DataStore.LOWEST_ALLOWED_VALUE;
		
		if (control < DataStore.LOWEST_ALLOWED_VALUE)
			control = DataStore.LOWEST_ALLOWED_VALUE;
		
		double [] dataValues = new double [] {signal,control,signal,control};
		data.put(p,dataValues);
	}

	public void addData (Probe p, double rawSignal, double rawControl, double normalisedSignal, double normalisedControl) {
		
		// We store the data in an array to save overhead.  The values
		// stored are:
		// 0 - raw signal
		// 1 - raw control
		// 2 - normalised signal
		// 3 - normalised control
		
		// Because we're often working with log data it's a real problem
		// if we get a signal or double value which is exactly 0.  We therefore
		// set a minimum value which these values can have.  We also avoid
		// divide by 0 errors when calculating ratios.
		
		if (rawSignal < DataStore.LOWEST_ALLOWED_VALUE)
			rawSignal = DataStore.LOWEST_ALLOWED_VALUE;
		
		if (rawControl < DataStore.LOWEST_ALLOWED_VALUE)
			rawControl = DataStore.LOWEST_ALLOWED_VALUE;

		if (normalisedSignal < DataStore.LOWEST_ALLOWED_VALUE)
			normalisedSignal = DataStore.LOWEST_ALLOWED_VALUE;
		
		if (normalisedControl < DataStore.LOWEST_ALLOWED_VALUE)
			normalisedControl = DataStore.LOWEST_ALLOWED_VALUE;

		
		double [] dataValues = new double [] {rawSignal,rawControl,normalisedSignal,normalisedControl};
		data.put(p,dataValues);
	}

	
	public void setSignalForProbe (Probe p, double value) {
		if (value < DataStore.LOWEST_ALLOWED_VALUE)
			value = DataStore.LOWEST_ALLOWED_VALUE;
		
		double [] d = (double [])data.get(p);
		if (d == null) {
			// For some reason we've not been given any data
			// for this probe.  We'll just subsitute fake
			// data to stop the calculations from screwing
			// up.
			addData(p,1,1);
			d = (double [])data.get(p);
		}

		d[2] = value;
	}

	public void setControlForProbe (Probe p, double value) {
		if (value < DataStore.LOWEST_ALLOWED_VALUE)
			value = DataStore.LOWEST_ALLOWED_VALUE;

		double [] d = (double [])data.get(p);
		if (d == null) {
			// For some reason we've not been given any data
			// for this probe.  We'll just subsitute fake
			// data to stop the calculations from screwing
			// up.
			addData(p,1,1);
			d = (double [])data.get(p);
		}
		d[3] = value;
	}

	public double getRawSignalValueForProbe (Probe p) {
		return ((double [])data.get(p))[0];
	}
	
	public double getRawControlValueForProbe (Probe p) {
		try {
			return ((double [])data.get(p))[1];
		}
		catch (Exception e) {
			System.out.println ("Probe is "+p+" value is "+data.get(p));
		}
		return -1;
	}

	public double getRawRatioForProbe (Probe p) {
		double [] d = (double [])data.get(p);
		return (d[0]/d[1]);
	}

	public double getSignalValueForProbe (Probe p) {
		return ((double [])data.get(p))[2];
	}
	
	public double getControlValueForProbe (Probe p) {
		return ((double [])data.get(p))[3];
	}

	public double getRatioForProbe (Probe p) {
		double [] d = (double [])data.get(p);
		return (d[2]/d[3]);
	}
	
	public String name () {
		return name;
	}
	
	public String toString () {
		return name();
	}

	public int compareTo(Object o) {
		return name.toLowerCase().compareTo(((DataSet)o).name.toLowerCase());
	}
}
