/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.GenomeParsers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkException;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Chromosome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Feature;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Genome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Location;
import uk.ac.bbsrc.babraham.ChipMonk.Preferences.ChipMonkPreferences;

public class EMBLParser implements Runnable {

	private Vector listeners = new Vector();
	private Genome genome;
	private File baseLocation;
	private int currentOffset = 0;
	private ChipMonkPreferences prefs = ChipMonkPreferences.getInstance();
	
	
	public void parseGenome (File baseLocation) {
		this.baseLocation = baseLocation;

		Thread t = new Thread(this);
		t.start();
	}
	
	public void addGenomeParserListener (GenomeParserListener gpl) {
		listeners.add(gpl);
	}
	
	public void removeGeneomeParserLister (GenomeParserListener gpl) {
		listeners.remove(gpl);
	}
		
	public void run() {
		try {
			genome = new Genome(baseLocation);
		} catch (ChipMonkException ex) {
			Enumeration en = listeners.elements();
			
			while (en.hasMoreElements()) {
				((GenomeParserListener)en.nextElement()).genomeExceptionReceived(ex);
				return;
			}
		}

		// We need a list of all of the .dat files inside the baseLocation
		File [] files = baseLocation.listFiles(new FileFilter() {
		
			public boolean accept(File f) {
				if (f.getName().toLowerCase().endsWith(".dat")) {
					return true;
				}
				else {
					return false;
				}
			}
		
		});
		
		for (int i=0;i<files.length;i++) {
			// Update the listeners
			Enumeration e = listeners.elements();
			
			while (e.hasMoreElements()) {
				((GenomeParserListener)e.nextElement()).genomeUpdateProgress(i,files.length,"Processing "+files[i].getName());
			}
			try {
				processFile(files[i]);
			} catch (Exception ex) {
				Enumeration en = listeners.elements();
				
				while (en.hasMoreElements()) {
					((GenomeParserListener)en.nextElement()).genomeExceptionReceived(ex);
				}
				return;
			}			
		}
		Enumeration en = listeners.elements();
		while (en.hasMoreElements()) {
			((GenomeParserListener)en.nextElement()).genomeProcessingFinished(genome);
		}

	}
	
	private void processFile (File f) throws Exception {
		
		BufferedReader br = new BufferedReader(new FileReader(f));

		Chromosome c = null;
		// We need to find and read the accession line to find out
		// which chromosome and location we're dealing with.
		
		// Each physical file can contain more than one EMBL file.  We 
		// need to account for this in our processing.
		
		while ((c = parseChromosome(br)) != null) {
			String line;			
			// We can now skip through to the start of the feature table
			while ((line=br.readLine())!=null) {
				if (line.startsWith("FH") || line.startsWith("SQ")) {
					break;
				}
			}
			
			// We can now start reading the features one at a time by
			// concatonating them and then passing them on for processing
			StringBuffer currentAttribute = new StringBuffer();
			boolean skipping = true;
			Feature feature = null;
			while ((line=br.readLine())!=null) {
				
//				System.err.println("Read line '"+line+"'");
				
				if (line.startsWith("XX") || line.startsWith("SQ") || line.startsWith("//")) {
					skipToEntryEnd(br);
					break;
				}
				
				if (line.length() < 17) continue;
				
				String type = line.substring(5,17).trim();
//				System.out.println("Type is "+type);
				if (type.length()>0) {
					//We're at the start of a new feature.
					
					// Check whether we need to process the old feature
					if (skipping) {
						// We're either on the first feature, or we've
						// moving past this one
						skipping = false;
					}
					else {						
						// We need to process the last attribute from the
						// old feature
						processAttributeReturnSkip(currentAttribute.toString(), feature);
						
						c.addFeature(feature);
					}
					
					// We can check to see if we're bothering to load this type of feature
					if (prefs.loadAnnotation(type)) {
//						System.err.println("Creating new feature of type "+type);
						feature = new Feature(type,c);
						currentAttribute=new StringBuffer("location=");
						currentAttribute.append(line.substring(21).trim());
						continue;
					}
					else {
//						System.err.println("Skipping feature of type "+type);
						genome.addUnloadedFeatureType(type);
						skipping = true;
					}
					
				}
				
				if (skipping) continue;
				
				String data = line.substring(21).trim();

				if (data.startsWith("/")) {
					// We're at the start of a new attribute
										
					//Process the last attribute
					skipping = processAttributeReturnSkip(currentAttribute.toString(), feature);
					currentAttribute = new StringBuffer();
				}
				
				// Our default action is just to append onto the existing information
				currentAttribute.append(data);
				
			}
			
			// We've finished, but we need to process the last feature
			// if there was one
			if (!skipping) {
				// We need to process the last attribute from the
				// old feature
				processAttributeReturnSkip(currentAttribute.toString(), feature);
				c.addFeature(feature);
			}

			
		}
	}
	
	private boolean processAttributeReturnSkip (String attribute, Feature feature) throws ChipMonkException {
//		System.out.println("Adding feature - current attribute is "+attribute);
		String [] nameValue = attribute.split("=",2);
		if (nameValue.length < 2) {
			// Some items are single values.  We'll fake a blank value
			String [] fakeValue = new String[2];
			fakeValue[0] = nameValue[0];
			fakeValue[1] = "";
			nameValue = fakeValue;
		}
		if (nameValue[0].equals("location")) {
//			System.out.println("Location is "+nameValue[1]);
			//Check to see if this is a location we can support
			
			if (nameValue[1].indexOf(":")>=0) {
				// Some locations are given relative to other sequences
				// (where a feature splits across more than one sequence).
				// We can't handle this so we don't try.
				return true;
			}

			feature.setLocation(new Location(nameValue[1],currentOffset));
		}
		else {
			// All other attributes just get added to the feature
//			System.out.println("Adding attribute '"+nameValue[0]+"' with value "+nameValue[1]);
			feature.addAttribute(nameValue[0], nameValue[1]);
			return false;
		}

		return false;
	}
	
	private Chromosome parseChromosome (BufferedReader br) throws ChipMonkException, IOException {
		String line;
		while ((line=br.readLine())!=null) {
			
			if (line.startsWith("AC")) {
				String [] sections = line.split(":");
				if (sections.length != 6) {
					// It's not a chromosome file.  We probably just want to
					// skip it and move onto the next entry
					skipToEntryEnd(br);
					continue;
				}
				if (line.indexOf("supercontig")>=0) {
					// It's not a chromosome file.  We probably just want to
					// skip it and move onto the next entry
					skipToEntryEnd(br);
					continue;
				}
				if (! genome.hasChromosome(sections[2])) {
					genome.addChromosome(sections[2]);
				}
				Chromosome c = genome.getChromosome(sections[2]);
								
				c.setLength(Integer.parseInt(sections[4]));
				
				// Since the positions of all features are given relative
				// to the current sequence we need to add the current
				// start position to all locations as an offset.
				currentOffset = Integer.parseInt(sections[3])-1;
				return c;
			}
			
			if (line.startsWith("//")) {
				throw new ChipMonkException("Couldn't find AC line");
			}
		}
		return null;
	}
	

	private void skipToEntryEnd (BufferedReader br) throws IOException {
		String line;
		try {
			Thread.sleep(5);
		} catch (InterruptedException e) {
		}

		while ((line=br.readLine())!=null) {
			if (line.startsWith("//"))
				return;
		}
	}
	
}
